#
# Copyright 2010 (C) Norwegian University of Science and Technology
#
# This file is part of Network Administration Visualized (NAV).
#
# NAV is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 2 as published by
# the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.  You should have received a copy of the GNU General Public License
# along with NAV. If not, see <http://www.gnu.org/licenses/>.
#
import time
from nav.Snmp.pysnmp_se import Snmp
from nav.Snmp.errors import *
from nav.bitvector import BitVector
from nav.models.manage import SwPortAllowedVlan

class SNMPHandler(object):
    netbox = ''
    ifAliasOid = '1.3.6.1.2.1.31.1.1.1.18' # From IF-MIB
    vlanOid = '1.3.6.1.2.1.17.7.1.4.5.1.1' # From Q-BRIDGE-MIB
    # oid for reading vlans om a stndard netbox (not cisco)
    dot1qVlanStaticRowStatus = '1.3.6.1.2.1.17.7.1.4.3.1.5'
    # List of all ports on a vlan as a hexstring
    dot1qVlanStaticEgressPorts = '1.3.6.1.2.1.17.7.1.4.3.1.2'
    ifAdminStatus = '1.3.6.1.2.1.2.2.1.7'
    ifOperStatus  = '1.3.6.1.2.1.2.2.1.8'

    def __init__(self, netbox):
        self.netbox = netbox
        self.readOnlyHandle = None
        self.readWriteHandle = None
    
    def __unicode__(self):
        return self.netbox.type.vendor.id

    def _bulkwalk(self, oid):
        handle = self._getReadOnlyHandle()
        result = []
        try:
            result = handle.bulkwalk(oid)
        except UnsupportedSnmpVersionError, e:
            result = handle.walk(oid)
        return result

    def _getLegalIfIndex(self, ifindex):
        if isinstance(ifindex, int):
            ifindex = str(ifindex)
        if not (isinstance(ifindex, str) or isinstance(ifindex, unicode)):
            raise TypeError('Illegal value for interface-index')
        if not ifindex.isdigit():
            raise TypeError('Illegal value for interface-index')
        return ifindex

    def _getQuery(self, oid, ifindex):
        return oid + "." + self._getLegalIfIndex(ifindex)

    def _getReadOnlyHandle(self):
        if self.readOnlyHandle is None:
            self.readOnlyHandle = Snmp(self.netbox.ip, self.netbox.read_only)
        return self.readOnlyHandle

    def _queryNetbox(self, oid, ifindex):
        handle = self._getReadOnlyHandle()
        result = None
        try:
            result = handle.get(self._getQuery(oid, ifindex))
        except NoSuchObjectError, e:
            pass
        return result

    def _getReadWriteHandle(self):
        if self.readWriteHandle is None:
            self.readWriteHandle = Snmp(self.netbox.ip,
                                        self.netbox.read_write,
                                        self.netbox.snmp_version)
        return self.readWriteHandle

    def _setNetboxValue(self, oid, ifindex, valueType, value):
        handle = self._getReadWriteHandle()
        return handle.set(self._getQuery(oid, ifindex), valueType, value)

    def getIfAlias(self, ifindex):
        """ Get alias on a specific interface """
        return self._queryNetbox(self.ifAliasOid, ifindex)

    def getAllIfAlias(self):
        return self._bulkwalk(self.ifAliasOid)
        
    def setIfAlias(self, ifindex, ifalias):
        """ Set alias on a specific interface."""
        if not (isinstance(ifalias, str) or isinstance(ifalias, unicode)):
            raise TypeError('Illegal value for interface-alias: %s' %ifalias)
        return self._setNetboxValue(self.ifAliasOid, ifindex, "s", ifalias)

    def getVlan(self, ifindex):
        """ Get vlan on a specific interface."""
        return self._queryNetbox(self.vlanOid, ifindex)

    def getAllVlans(self):
        return self._bulkwalk(self.vlanOid)
    
    def _computeOctetString(self, hexstring, port, action='enable'):
        """
        hexstring: the returnvalue of the snmpquery
        port: the number of the port to add
        """
        bit = BitVector(hexstring)
        # Add port to string
        port = port - 1
        if action == 'enable':
            bit[port] = 1
        else:
            bit[port] = 0
        return str(bit)

    def setVlan(self, ifindex, vlan):
        if isinstance(vlan, str) or isinstance(vlan, unicode):
            if vlan.isdigit():
                vlan = int(vlan)
        if not isinstance(vlan, int):
            raise TypeError('Illegal value for vlan: %s' %vlan)
        # Fetch current vlan
        fromvlan = self.getVlan(ifindex)
        # fromvlan and vlan is the same, there's nothing to do
        if fromvlan == vlan:
            return None
        # Add port to vlan. This makes the port active on both old and new vlan
        status = self._setNetboxValue(self.vlanOid, ifindex, "u", vlan)
        # Remove port from list of ports on old vlan
        hexstring = self._queryNetbox(self.dot1qVlanStaticEgressPorts, fromvlan)
        modified_hexport = self._computeOctetString(hexstring, ifindex, 'disable')
        return self._setNetboxValue(self.dot1qVlanStaticEgressPorts, fromvlan, 's', modified_hexport)


    def setIfUp(self, ifindex):
        """Set interface.to up"""
        return self._setNetboxValue(self.ifAdminStatus, ifindex, "i", 1)

    def setIfDown(self, ifindex):
        """Set interface.to down"""
        return self._setNetboxValue(self.ifAdminStatus, ifindex, "i", 2)

    def restartIf(self, ifindex, wait=5):
        """ Take interface down and up.
            wait = number of seconds to wait between down and up."""
        if isinstance(wait, str) or isinstance(wait, unicode):
            if wait.isdigit():
                wait = int(wait)
        if not isinstance(wait, int):
            raise TypeError('Illegal value for wait: %s' %wait)
        self.setIfDown(ifindex)
        time.sleep(wait)
        self.setIfUp(ifindex)

    def write_mem(self):
        """ 
        Do a write memory on netbox if available
        """
        pass

    def getIfAdminStatus(self, ifindex):
        return self._queryNetbox(self.ifAdminStatus, ifindex)

    def getIfOperStatus(self, ifindex):
        return self._queryNetbox(self.ifOperStatus, ifindex)

    def _getLastNumber(self, oid):
        if not (isinstance(oid, str) or isinstance(oid, unicode)):
            raise TypeError('Illegal value for oid')
        splits = oid.split('.')
        last = splits[-1]
        if isinstance(last, str):
            if last.isdigit():
                last = int(last)
        return last

    def _getIfStats(self, stats):
        available_stats = []
        for (ifIndex, stat) in stats:
            ifIndex = self._getLastNumber(ifIndex)
            if isinstance(ifIndex, int):
                available_stats.append((ifIndex, stat))
        return available_stats
            
    def getNetboxAdminStatus(self):
        ifAdminStats = self._bulkwalk(self.ifAdminStatus)
        return self._getIfStats(ifAdminStats)

    def getNetboxOperStatus(self):
        ifOperStats = self._bulkwalk(self.ifOperStatus)
        return self._getIfStats(ifOperStats)

    def _filter_vlans(self, vlans):
        vlans = filter(None, list(set(vlans)))
        vlans.sort()
        return vlans

    def getNetboxVlans(self):
        boxVlans = self._bulkwalk(self.dot1qVlanStaticRowStatus)
        available_vlans = []
        for (vlan, valueType) in boxVlans:
            currVlan = self._getLastNumber(vlan)
            if isinstance(currVlan, int):
                available_vlans.append(currVlan)
        # remove duplicates and none values
        available_vlans = self._filter_vlans(available_vlans)
        return available_vlans

class Cisco(SNMPHandler):
    def __init__(self, netbox):
        super(Cisco, self).__init__(netbox)
        self.vlanOid = '1.3.6.1.4.1.9.9.68.1.2.2.1.2'
        self.writeMemOid = '1.3.6.1.4.1.9.2.1.54.0'

    def getNetboxVlans(self):
        """ Find all available vlans on this netbox"""
        available_vlans = []
        for swport in self.netbox.get_swports():
            if swport.trunk:
                available_vlans.extend(self._find_vlans_on_trunk(swport))
            else:
                available_vlans.append(swport.vlan)
        # remove duplicates and none values
        available_vlans = self._filter_vlans(available_vlans)
        return available_vlans

    def _find_vlans_on_trunk(self, swport):
        """Find all vlans on trunk-ports"""
        port = SwPortAllowedVlan.objects.get(interface=swport.id)
        vector = BitVector.from_hex(port.hex_string)
        vlans = vector.get_set_bits()
        return vlans

    def setVlan(self, ifindex, vlan):
        if isinstance(vlan, str) or isinstance(vlan, unicode):
            if vlan.isdigit():
                vlan = int(vlan)
        if not isinstance(vlan, int):
            raise TypeError('Illegal value for vlan: %s' %vlan)
        # Fetch current vlan
        fromvlan = self.getVlan(ifindex)
        # fromvlan and vlan is the same, there's nothing to do
        if fromvlan == vlan:
            return None
        # Add port to vlan. This makes the port active on both old and new vlan
        status = None
        try:
            status = self._setNetboxValue(self.vlanOid, ifindex, "u", vlan)
        except Exception, e:
            # Ignore this exception,- some boxes want signed integer and
            # we do not know this beforehand.
            # If unsigned fail,- try with signed integer.
            status = self._setNetboxValue(self.vlanOid, ifindex, "i", vlan)
        return status
            
   
    def write_mem(self):
        """ Use OLD-CISCO-SYS-MIB (v1) writeMem to write to memory. 
        Write configuration into non-volatile memory / erase config memory if 0.
        """
        handle = self._getReadWriteHandle()
        return handle.set(self.writeMemOid, 'i', 1)
        
        
class HP(SNMPHandler):
    def __init__(self, netbox):
        super(HP, self).__init__(netbox)

VENDOR_CISCO = 9
VENDOR_HP = 11

class SNMPFactory(object):
    @classmethod
    def getInstance(self, netbox):
        vendor_id = netbox.type.get_enterprise_id()
        if (vendor_id == VENDOR_CISCO):
            return Cisco(netbox)
        if (vendor_id == VENDOR_HP):
            return HP(netbox)
        return SNMPHandler(netbox)

    def __init__(self):
        pass
